# Author: Toby Dickenson, <htrd90@zepler.org>
#
# Copyright (c) 1999 Toby Dickenson
#
# Permission to use this software in any way is granted without
# fee, provided that the copyright notice above appears in all
# copies. This software is provided "as is" without any warranty.
#

"""NT User folder product

Authenticates off NT security as applied to a registry key

"""

from Acquisition import Implicit
from ImageFile import ImageFile
from AccessControl.Role import RoleManager
from AccessControl.User import BasicUser, BasicUserFolder
from Globals import HTMLFile, Persistent
from App.Dialogs import MessageDialog
import win32api
import win32con
import win32security
import string
import Threading
import traceback
import time
import win32net, win32netcon

key_prefix="SOFTWARE\\Digital Creations\\zope\\Users"

_users_root=win32api.RegCreateKey(win32con.HKEY_LOCAL_MACHINE,key_prefix)

maximum_number_of_roles=100

class NTUserFolder(BasicUserFolder):
    """User folder for NT security"""

    meta_type='NTUserFolder'
    id       ='acl_users'
    title    ='NTUserFolder'
    icon     ='misc_/NTUserFolder/nt_icon'

    manage_options=(
    {'label':'Contents',   'action':'manage_main'},
    {'label':'Security',   'action':'manage_access'},
    {'label':'Undo',       'action':'manage_UndoForm'},
    )

    __ac_permissions__=BasicUserFolder.__ac_permissions__+ (
    ('Manage NT User Folder',   ['manage_edit']),   )

    manage_main=HTMLFile('mainUser', globals())

    def __init__(self, key):
        self.key=key

    def manage_edit(self,key,REQUEST):
        "For changing key in management screen"
        if not valid_key(key):  return MessageDialog(
                                           title  ='Bad Key',
                                           message='That key is invalid',
                                           action ='%s/manage_main' % REQUEST['PARENT_URL'])
        self.key=key
        return MessageDialog(
                   title  ='Key Changed',
                   message='That key has been changed',
                   action ='%s/manage_main' % REQUEST['PARENT_URL'])

    def getUserNames(self):
        "Returns a list of user names"
        names = []
        resume = 0
        while 1:
            users,junk,resume = win32net.NetUserEnum(None,10,win32netcon.FILTER_NORMAL_ACCOUNT,resume)
            names = names+map(lambda user: user['name'], users)
            if not resume:
                break
        return names

    def getUsers(self):
        "Return a list of user objects"
        # TODO: see where this is used, how essential is it
        # TODO: possibly enumerate all NT users
        return []

    def getUser(self, name):
        "Return the named user object or None if no such user exists"
        try:
            info=win32net.NetUserGetInfo(None,name,2)
            name=str(info['name'])
            full_name=str(info['full_name'])
            description=str(info['comment'])
        except win32api.error:
            return None
        return NTUser(name,full_name,description,key_prefix+"\\"+self.key)

    def defined_roles(self):
        "A list of the roles that are defined by this key"
        roles=[]
        hkey=win32api.RegOpenKey(_users_root,self.key)
        try:
            for i in range(maximum_number_of_roles):
                try:
                    roles.append(win32api.RegEnumKey(hkey,i))
                except win32api.error:
                    # No more subkeys
                    break
        finally:
            win32api.RegCloseKey(hkey)
        return roles

    def list_sub(self,a,b):
        "Find all the elements of list a not in b"
        return filter(lambda element, b=b[:]: not(element in b) , a[:])

class NTUser(BasicUser):
    "An object representing an NT user"

    def __init__(self,name,full_name,description,key):
        self._name=name
        self._full_name=full_name
        self._description=description
        self._thread=None
        self._key=key

    def getUserName(self):
        "Return the username of a user"
        return self._name

    def getFullUserName(self):
        "Return the full name of a user"
        return self._full_name

    def getDescription(self):
        "Return the description of a user"
        # TODO: why does this sometimes not work?
        return self._description

    def getRoles(self):
        "Return the list of roles assigned to a user."
        return self._thread.get_roles()

    def getTimeToAuthenticate(self):
        "Return the time taken to perform authentication."
        return self._thread.get_elapsed_time()

    def authenticate(self, password, request):
        self._thread=authentication_thread(self._name,password,self._key)
        self._thread.start()
        # Authentication runs in a separate thread since it uses Impersonation. We want to
        # run the minimum amount of code under the credentials of the user we are authenticating.
        return self._thread.is_authentic()



class authentication_thread(Threading.Thread):
    """Class representing the thread which performs the authentication. Instances of this class
    are not thread-safe; the public members can only be used by the thread that creates it."""
    
    def __init__(self,name,password,full_key):
        Threading.Thread.__init__(self)
        self.name=name
        self.password=password
        self.full_key=full_key
        self._authenticated=0
        self._roles=[]
        self._authentication_lock=Threading.Lock()
        self._authentication_lock.acquire()
        self._role_lock=Threading.Lock()
        self._role_lock.acquire()

    def run(self):
        self._token=None
        self._hkey=None
        try:
            start=time.clock()
            try:
                try:
                    self._authenticate()
                finally:
                    self._authentication_lock.release()

                if self._authenticated:
                    self._get_roles()
            finally:
                self._elapsed_time=time.clock()-start
                self._role_lock.release()

        finally:
            if self._token: win32api.CloseHandle(self._token)
            if self._hkey: win32api.RegCloseKey(self._hkey)

    def _authenticate(self):
        try:
            # User must be able to log in
            self._token=win32security.LogonUser(self.name,".",self.password,win32security.LOGON32_LOGON_NETWORK,win32security.LOGON32_PROVIDER_DEFAULT)
            win32security.ImpersonateLoggedOnUser(self._token)
            self._hkey=win32api.RegOpenKey(win32con.HKEY_LOCAL_MACHINE,self.full_key)
            # User must be able to read from the controlling key
            win32api.RegQueryValue(self._hkey,None)
            self._authenticated=1
        except win32api.error:
            pass

    def _get_roles(self):
        for i in range(maximum_number_of_roles):
            try:
                role_name=win32api.RegEnumKey(self._hkey,i)
            except win32api.error:
                # No more subkeys
                return
            try:
                win32api.RegQueryValue(self._hkey,role_name)
                self._roles.append(role_name)
            except win32api.error:
                # User does not have this role
                pass

    def is_authentic(self):
        if self._authentication_lock:
            # The first time anyone has asked... wait for completion
            self._authentication_lock.acquire()
            self._authentication_lock=None
        return self._authenticated
            
    def get_roles(self):
        if self._role_lock:
            # The first time anyone has asked... wait for completion
            self._role_lock.acquire()
            self._role_lock=None
        return tuple(self._roles)

    def get_elapsed_time(self):            
        self.get_roles()
        return self._elapsed_time


# Probably over cautious
safe_characters=string.uppercase+string.lowercase+string.digits

def valid_key(key):
    # Must be sensble length
    if not 1<len(key)<50:
        return 0
    # And alphanumeric
    for c in str(key):
        if not c in safe_characters:
            return 0
    # And createable or already created
    try: hkey=win32api.RegCreateKey(_users_root,key)
    except: return 0
    win32api.RegCloseKey(hkey)
    return 1
    


manage_addNTUserFolderForm=HTMLFile('addNTUserFolder', globals())

def manage_addNTUserFolder(self, key, REQUEST):
    """Add NT User folder"""

    if not valid_key(key):
        return MessageDialog(
                   title  ='Bad Key',
                   message='That key is invalid',
                   action ='%s/manage_main' % REQUEST['PARENT_URL'])

    try:    self._setObject('acl_users', NTUserFolder(key))
    except: return MessageDialog(
                   title  ='Item Exists',
                   message='This object already contains a User Folder',
                   action ='%s/manage_main' % REQUEST['PARENT_URL'])

    self.__allow_groups__=self.acl_users

    return self.manage_main(self,REQUEST,update_menu=1)

